# https://github.com/nonebot/nonebot2/blob/master/tests/utils.py

from collections.abc import Iterable, Mapping
from typing import Optional, Union
from typing_extensions import override

from nonebot.adapters import Event, Message, MessageSegment
from pydantic import create_model


class FakeMessageSegment(MessageSegment["FakeMessage"]):
    @classmethod
    @override
    def get_message_class(cls):
        return FakeMessage

    @override
    def __str__(self) -> str:
        return self.data["text"] if self.type == "text" else f"[fake:{self.type}]"

    @classmethod
    def text(cls, text: str):
        return cls("text", {"text": text})

    @staticmethod
    def image(url: str):
        return FakeMessageSegment("image", {"url": url})

    @staticmethod
    def nested(content: "FakeMessage"):
        return FakeMessageSegment("node", {"content": content})

    @override
    def is_text(self) -> bool:
        return self.type == "text"


class FakeMessage(Message[FakeMessageSegment]):
    @classmethod
    @override
    def get_segment_class(cls):
        return FakeMessageSegment

    @staticmethod
    @override
    def _construct(msg: Union[str, Iterable[Mapping]]):
        if isinstance(msg, str):
            yield FakeMessageSegment.text(msg)
        else:
            for seg in msg:
                yield FakeMessageSegment(**seg)


def make_fake_event(
    _base: Optional[type[Event]] = None,
    _type: str = "message",
    _name: str = "test",
    _description: str = "test",
    _user_id: Optional[str] = "test",
    _session_id: Optional[str] = "test",
    _message: Optional[Message] = None,
    _to_me: bool = True,
    **fields,
) -> type[Event]:
    Base = _base or Event  # noqa: N806

    class FakeEvent(Base):
        @override
        def get_type(self) -> str:
            return _type

        @override
        def get_event_name(self) -> str:
            return _name

        @override
        def get_event_description(self) -> str:
            return _description

        @override
        def get_user_id(self) -> str:
            if _user_id is not None:
                return _user_id
            raise NotImplementedError

        @override
        def get_session_id(self) -> str:
            if _session_id is not None:
                return _session_id
            raise NotImplementedError

        @override
        def get_message(self) -> "Message":
            if _message is not None:
                return _message
            raise NotImplementedError

        @override
        def is_tome(self) -> bool:
            return _to_me

    return create_model("FakeEvent", __base__=FakeEvent, **fields)



from typing import TYPE_CHECKING, Callable

from ..common import (
    NameDecoCollector,
    camel_case,
    escape_backticks,
    escape_double_quotes,
    escape_single_quotes,
    full_to_half,
)

if TYPE_CHECKING:
    from jinja2 import Environment


cookit_global_filter = NameDecoCollector[Callable]()

all_filters = cookit_global_filter.data
"""@deprecated: Use `cookit_global_filter.data` instead."""


cookit_global_filter(camel_case)
cookit_global_filter(full_to_half)
cookit_global_filter(escape_backticks)
cookit_global_filter(escape_double_quotes)
cookit_global_filter(escape_single_quotes)


@cookit_global_filter
def br(value: str) -> str:
    return value.replace("\n", "<br/>")


def register_all_filters(env: "Environment"):
    env.filters.update(cookit_global_filter.data)



import json
import time
from pathlib import Path
from typing import Any


class DebugFileWriter:
    def __init__(self, base_path: Path, *rest_path: str) -> None:
        self.base_path = base_path
        self.full_path = base_path.joinpath(*rest_path)

    @property
    def enabled(self) -> bool:
        return self.base_path.exists()

    def write(self, content: Any, filename: str):
        filename = filename.format(time=round(time.time() * 1000))
        self.full_path.mkdir(parents=True, exist_ok=True)
        path = self.full_path / filename
        if isinstance(content, (bytes, bytearray)):
            path.write_bytes(content)
        else:
            path.write_text(
                (
                    content
                    if isinstance(content, str)
                    else json.dumps(content, ensure_ascii=False)
                ),
                "u8",
            )



import time
from typing import TYPE_CHECKING, Any
from typing_extensions import override

if TYPE_CHECKING:
    from nonebot.adapters import Bot as BaseBot
    from nonebot.adapters.satori import Adapter as SatoriAdapter
    from nonebug import App
    from nonebug.mixin.process import MatcherContext


def create_fake_satori_adapter(ctx: "MatcherContext") -> "SatoriAdapter":
    from nonebot import get_driver
    from nonebot.adapters.satori import Adapter as SatoriAdapter

    class FakeAdapter(SatoriAdapter):
        @override
        async def _call_api(self, bot: "BaseBot", api: str, **data) -> Any:
            return ctx.got_call_api(self, api, **data)

    return FakeAdapter(get_driver())


async def test_captured_recall(app: "App"):
    from nonebot import on_message, require
    from nonebot.adapters.satori import Bot as SatoriBot, Message, MessageSegment
    from nonebot.adapters.satori.config import ClientInfo
    from nonebot.adapters.satori.event import MessageCreatedEvent
    from nonebot.adapters.satori.models import (
        ChannelType,
        LoginOnline,
        LoginStatus,
        MessageObject,
        User,
    )
    from nonebot.compat import type_validate_python

    require("nonebot_plugin_alconna")

    from cookit.nonebot.alconna import RecallContext
    from nonebot_plugin_alconna.uniseg import UniMessage

    matcher = on_message()

    @matcher.handle()
    async def _():
        async with RecallContext(delay=0, wait=True) as recall:
            await recall.send(UniMessage("world!"))
            recall.append(await UniMessage("world!").send())

    async with app.test_matcher(matcher) as ctx:
        adapter = create_fake_satori_adapter(ctx)
        login = LoginOnline(
            sn=0,
            adapter="fake",
            user=User(id="fake_bot"),
            platform="fake",
            status=LoginStatus.ONLINE,
        )
        bot = ctx.create_bot(
            base=SatoriBot,
            adapter=adapter,
            self_id="fake_bot",
            login=login,
            info=ClientInfo(port=14514),
            proxy_urls=[],
        )
        event = type_validate_python(
            MessageCreatedEvent,
            {
                "sn": 1,
                "type": "message-created",
                "login": login,
                "timestamp": int(time.time() * 1000),
                "channel": {"id": "fake_channel", "type": ChannelType.TEXT},
                "user": {"id": "fake_user"},
                "message": {"id": "fake_message", "content": "hello"},
            },
        )
        ctx.receive_event(bot, event)

        # ctx.should_call_api(
        #     "message_create",
        #     {"channel_id": "fake_channel", "content": "world!"},
        #     [MessageObject(id="114515", content="world!")],
        # )
        # ctx.should_call_api(
        #     "message_create",
        #     {"channel_id": "fake_channel", "content": "world!"},
        #     [MessageObject(id="114516", content="world!")],
        # )
        ctx.should_call_send(
            event,
            Message() + MessageSegment.text("world!"),
            [MessageObject(id="114515", content="world!")],
        )
        ctx.should_call_send(
            event,
            Message() + MessageSegment.text("world!"),
            [MessageObject(id="114516", content="world!")],
        )

        ctx.should_call_api(
            "message_delete",
            {"channel_id": "fake_channel", "message_id": "114515"},
        )
        ctx.should_call_api(
            "message_delete",
            {"channel_id": "fake_channel", "message_id": "114516"},
        )



from pathlib import Path

import jinja2

from ...utils import expect_exc, mark_test
from ..utils import get_page


@mark_test
async def test_template_renderer():
    from cookit.pw.jinja import make_template_renderer

    template_text = (Path(__file__).parent / "test.html.jinja").read_text("u8")
    template = jinja2.Template(template_text, autoescape=True, enable_async=True)
    renderer = await make_template_renderer(template, title_tag="h1")

    async with await get_page() as page:
        await renderer(page)

    async with await get_page() as page:
        await renderer(page, "h1")

    async with await get_page() as page:
        with expect_exc(ValueError):
            await renderer(page, "h5")



from typing import TYPE_CHECKING, Optional
from typing_extensions import Unpack

from loguru import logger

if TYPE_CHECKING:
    from ..router import CKRouterFunc, CKRouterKwArgs


def log_router_err(error_code: Optional[str] = None):
    def deco(router: "CKRouterFunc") -> "CKRouterFunc":
        async def wrapper(**kwargs: Unpack["CKRouterKwArgs"]):
            try:
                return await router(**kwargs)
            except Exception:
                logger.exception("Error occurred when handling route")
                await kwargs["route"].abort(error_code)

        return wrapper

    return deco



from nonebot import get_plugin_config
from nonebot_plugin_localstore import plugin_config as localstore_config
from pydantic import BaseModel, Field


class LocalStoreConfigModel(BaseModel):
    sure: bool = Field(
        default=False,
        alias="i_am_sure_i_want_to_use_default_localstore_file_path",
    )


LOCALSTORE_DEFAULT_PATH_WARNING_MSG = (
    "\nWarning:"
    "\n"
    "\nDetected that the `localstore` plugin is using default storage locations:"
    "\n- On Windows:"
    "\n    C:\\Users\\<username>\\AppData\\Local (and Roaming)\\nonebot2"
    "\n- On Linux:"
    "\n    ~/.cache/nonebot2, ~/.config/nonebot2, and ~/.local/share/nonebot2"
    "\n- On macOS:"
    "\n    ~/Library/Caches/nonebot2 and ~/Library/Application Support/nonebot2"
    "\n"
    "\nI strongly dislike this default storage approach as it:"
    "\n- Increases complexity when migrating NoneBot instances"
    "\n- May cause data conflicts when running multiple NoneBot instances simultaneously"
    "\nTherefore, this plugin will REFUSE TO LOAD."
    "\n"
    "\nRecommend to add this configuration"
    " to let localstore plugin store data in the current working directory:"
    "\n- LOCALSTORE_USE_CWD=True"
    "\nOr edit your configuration file according to localstore plugin's documentation"
    " to configure custom storage paths"
    "\n"
    "\nIf you insist on using default storage paths,"
    " add this configuration to ignore this error:"
    "\n- I_AM_SURE_I_WANT_TO_USE_DEFAULT_LOCALSTORE_FILE_PATH=True"
    "\n"
    "\n警告："
    "\n"
    "\n检测到 `localstore` 插件正在使用默认的文件存放位置，即："
    "\n- 在 Windows 系统下："
    "\n    C:\\Users\\<username>\\AppData\\Local (及 Roaming)\\nonebot2"
    "\n- 在 Linux 系统下："
    "\n    ~/.cache/nonebot2 与 ~/.config/nonebot2 与 ~/.local/share/nonebot2"
    "\n- 在 macOS 系统下："
    "\n    ~/Library/Caches/nonebot2 与 ~/Library/Application Support/nonebot2"
    "\n"
    "\n由于本人极不喜欢 localstore 插件这样存放数据文件，"
    "\n这不仅增加了用户迁移 NoneBot 实例的复杂程度，"
    "\n而且在同时运行多个 NoneBot 实例的情况下还可能会导致实例间数据冲突。"
    "\n因此，本插件将 拒绝加载。"
    "\n"
    "\n建议添加下方配置项使 localstore 插件在当前工作目录下存储数据："
    "\n- LOCALSTORE_USE_CWD=True"
    "\n或者参考其文档自行修改配置使用其他文件夹作为数据存储位置。"
    "\n"
    "\n如果你坚决要使用 localstore 插件的默认文件存放位置，请添加以下配置项以忽略本错误："
    "\n- I_AM_SURE_I_WANT_TO_USE_DEFAULT_LOCALSTORE_FILE_PATH=True"
)


def ensure_localstore_path_config():
    if (
        (not localstore_config.localstore_use_cwd)
        and (not localstore_config.localstore_cache_dir)
        and (not localstore_config.localstore_config_dir)
        and (not localstore_config.localstore_data_dir)
        and (not get_plugin_config(LocalStoreConfigModel).sure)
    ):
        raise RuntimeError(LOCALSTORE_DEFAULT_PATH_WARNING_MSG)



import base64
from collections.abc import Container, Iterable, Iterator, Sequence
from contextlib import suppress
from typing import Any, Callable, Optional, TypeVar, Union, cast, overload
from typing_extensions import ParamSpec, TypeGuard

T = TypeVar("T")

TA = TypeVar("TA")
TB = TypeVar("TB")
N = TypeVar("N")

K = TypeVar("K")
V = TypeVar("V")

P = ParamSpec("P")

LazyGetterType = Union[T, Callable[P, T]]


def lazy_get(
    val: "LazyGetterType[T, P]",
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    return cast("T", val(*args, **kwargs) if callable(val) else val)


def qor(
    a: Union[TA, N],
    b: "LazyGetterType[TB, []]",
    none_val: N = None,
) -> Union[TA, TB]:
    def guard(x: Union[TA, N]) -> TypeGuard[TA]:
        return x is not none_val

    return a if guard(a) else lazy_get(b)


def chunks(lst: Sequence[T], n: int) -> Iterator[list[T]]:
    for i in range(0, len(lst), n):
        yield list(lst[i : i + n])


def flatten(li: Iterable[Iterable[T]]) -> list[T]:
    return [x for y in li for x in y]


def set_default(target: dict[K, V], key: K, default: "LazyGetterType[V, []]") -> V:
    if key in target:
        return target[key]
    default = lazy_get(default)
    target[key] = default
    return default


@overload
def auto_delete(target: dict[K, V], transform: None = None) -> dict[K, V]: ...
@overload
def auto_delete(
    target: dict[K, V],
    transform: Callable[[V], Optional[T]],
) -> dict[K, T]: ...
def auto_delete(  # noqa: E302
    target: dict[K, V],
    transform: Optional[Callable[[V], Optional[T]]] = None,
) -> dict[K, Any]:
    data: dict[K, Union[V, T]] = {}
    for k, v in tuple(target.items()):
        vt = transform(v) if transform else v
        if vt:
            data[k] = vt
        else:  # expected behavior, bool(vt) == False will be removed
            del target[k]
    return data


def to_b64_url(data: bytes, mime: Optional[str] = None) -> str:
    if mime is None:
        mime = ""
        with suppress(ImportError, IndexError):
            import fleep

            mime = cast("list[str]", fleep.get(data[:128]).mime)[0]

    return f"data:{mime};base64,{base64.b64encode(data).decode()}"


def deep_merge(
    *args_input: Any,
    skip_merge_paths: Optional[Container[str]] = None,
) -> Any:
    def merge_path(current: Optional[str], *paths: str) -> str:
        x = []
        if current:
            x.append(current)
        x.extend(paths)
        return ".".join(x)

    def merge_obj(*args: Any, current_path: Optional[str] = None) -> Any:
        if skip_merge_paths and (current_path in skip_merge_paths):
            return args[-1]

        if all(isinstance(x, dict) for x in args):
            keys = {x for d in args for x in d}
            result = {}
            for k in keys:
                result[k] = merge_obj(
                    *(d[k] for d in args if k in d),
                    current_path=merge_path(current_path, k),
                )
            return result

        if issubclass((seq_type := type(args[0])), (list, set)) and all(
            isinstance(x, seq_type) for x in args
        ):
            return seq_type(x for ls in args for x in ls)

        return args[-1]

    return merge_obj(*args_input)



"""Install `cookit[nonebot-localstore]` before import this module."""

# region assert required plugin loaded
from .. import assert_plugin_loaded

assert_plugin_loaded("nonebot_plugin_localstore")
del assert_plugin_loaded
# endregion

# ruff: noqa: E402
from .base import (
    LOCALSTORE_DEFAULT_PATH_WARNING_MSG as LOCALSTORE_DEFAULT_PATH_WARNING_MSG,
    ensure_localstore_path_config as ensure_localstore_path_config,
)



import pytest
from pytest_asyncio import is_async_test


def pytest_configure(config: pytest.Config) -> None:
    from nonebug import NONEBOT_INIT_KWARGS

    config.stash[NONEBOT_INIT_KWARGS] = {
        "driver": "~fastapi+~websockets+~httpx",
        "log_level": "DEBUG",
    }


def pytest_collection_modifyitems(items: list[pytest.Item]):
    pytest_asyncio_tests = (item for item in items if is_async_test(item))
    session_scope_marker = pytest.mark.asyncio(loop_scope="session")
    for async_test in pytest_asyncio_tests:
        async_test.add_marker(session_scope_marker, append=False)


@pytest.fixture(scope="session", autouse=True)
def load_adapters(nonebug_init: None):  # noqa: ARG001, PT004
    from nonebot import get_driver
    from nonebot.adapters.satori import Adapter as SatoriAdapter

    driver = get_driver()
    driver.register_adapter(SatoriAdapter)



import re
from pathlib import Path
from typing import TYPE_CHECKING, Optional, cast

from ..utils import mark_test
from .utils import get_page

if TYPE_CHECKING:
    from playwright.async_api import Route
    from yarl import URL

ROUTE_BASE_URL = "https://cookit.route"


@mark_test
async def test_router_group_route():
    from cookit.pw import RouterGroup

    router_group = RouterGroup()

    body_root = "root"

    @router_group.router(f"{ROUTE_BASE_URL}/")
    async def _(route: "Route", **_):
        await route.fulfill(body=body_root)

    body_test = "test"

    @router_group.router(f"{ROUTE_BASE_URL}/test/**/*")
    @router_group.router(f"{ROUTE_BASE_URL}/test")
    async def _(route: "Route", **_):
        await route.fulfill(body=body_test)

    body_test2_re = "test2_re"

    last_matched = cast("Optional[re.Match[str]]", None)

    @router_group.router(re.compile(rf"{ROUTE_BASE_URL}/test2(/.*)?"))
    async def _(route: "Route", matched: Optional[re.Match[str]], **_):
        nonlocal last_matched
        last_matched = matched
        await route.fulfill(body=body_test2_re)

    async with await get_page() as page:
        await router_group.apply(page)

        resp = await page.goto(f"{ROUTE_BASE_URL}/")
        assert resp
        assert (await resp.body()).decode() == body_root

        resp = await page.goto(f"{ROUTE_BASE_URL}/test")
        assert resp
        assert (await resp.body()).decode() == body_test

        resp = await page.goto(f"{ROUTE_BASE_URL}/test/test/test")
        assert resp
        assert (await resp.body()).decode() == body_test

        resp = await page.goto(f"{ROUTE_BASE_URL}/test2")
        assert resp
        assert (await resp.body()).decode() == body_test2_re
        assert last_matched
        assert not last_matched[1]

        resp = await page.goto(f"{ROUTE_BASE_URL}/test2/test")
        assert resp
        assert (await resp.body()).decode() == body_test2_re
        assert last_matched
        assert last_matched[1] == "/test"


@mark_test
async def test_real_path_router():
    from cookit.pw import RouterGroup, make_real_path_router

    router_group = RouterGroup()

    base_path = Path(__file__).parent

    @router_group.router(f"{ROUTE_BASE_URL}/**/*")
    @make_real_path_router
    async def _(url: "URL", **_):
        return base_path.joinpath(url.path[1:])

    async with await get_page() as page:
        await router_group.apply(page)

        resp = await page.goto(f"{ROUTE_BASE_URL}/test.html")
        assert resp
        assert (await resp.body()) == (base_path / "test.html").read_bytes()

        resp = await page.goto(f"{ROUTE_BASE_URL}/unknown_file.html")
        assert resp
        assert resp.status == 404


@mark_test
async def test_router_group_copy():
    from cookit.pw import RouterGroup

    router_group = RouterGroup()

    @router_group.router(f"{ROUTE_BASE_URL}/test")
    async def _(route: "Route", **_):
        await route.fulfill(body=b"")

    router_group2 = router_group.copy()

    @router_group2.router(f"{ROUTE_BASE_URL}/test2")
    async def _(route: "Route", **_):
        await route.fulfill(body=b"")

    assert router_group.routers is not router_group2.routers
    assert len(router_group.routers) == 1
    assert len(router_group2.routers) == 2



from typing import TYPE_CHECKING, Optional

from ..screenshot import screenshot_html

if TYPE_CHECKING:
    from jinja2 import Template
    from playwright.async_api import Page


async def make_template_renderer(template: "Template", **template_kwargs):
    html = await template.render_async(**template_kwargs)

    async def render(page: "Page", selector: Optional[str] = None, **screenshot_kwargs):
        return await screenshot_html(page, html, selector, **screenshot_kwargs)

    return render



from typing import Optional

from playwright.async_api import Browser, Playwright, async_playwright

from tests_standalone.utils import mark_before_exit

pw: Optional[Playwright] = None
browser: Optional[Browser] = None


async def get_playwright():
    global pw
    if not pw:
        pw = await async_playwright().start()
    return pw


async def get_browser():
    global browser
    if not browser:
        browser = await (await get_playwright()).chromium.launch()
    return browser


async def get_page():
    return await (await get_browser()).new_page()


@mark_before_exit
async def _():
    if browser:
        await browser.close()
    if pw:
        await pw.stop()



import asyncio
import importlib
import sys
from asyncio import Semaphore
from collections.abc import Awaitable, Coroutine
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, Union
from typing_extensions import ParamSpec, override

if TYPE_CHECKING:
    from types import ModuleType

T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")
P2 = ParamSpec("P2")


def with_semaphore(semaphore: Semaphore):
    def decorator(func: Callable[P, Awaitable[R]]):
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            async with semaphore:
                return await func(*args, **kwargs)

        return wrapper

    return decorator


def queued(func: Callable[P, Awaitable[R]]):
    return with_semaphore(Semaphore(1))(func)


async def race(*coros: Coroutine[Any, Any, T]) -> T:
    done, pending = await asyncio.wait(
        [asyncio.create_task(x) for x in coros],
        return_when=asyncio.FIRST_COMPLETED,
    )
    first, *other = done
    for t in (*other, *pending):
        t.cancel()
    return first.result()


def auto_import(
    path: Union[str, Path],
    package: Optional[str] = None,
    path_filter: Optional[Callable[[Path], bool]] = None,
):
    if not isinstance(path, Path):
        path = Path(path)
    if not path_filter:
        path_filter = lambda x: not x.name.startswith("_")  # noqa: E731

    modules: list[ModuleType] = []
    for p in path.iterdir():
        if (
            (not (p / "__init__.py").exists()) if p.is_dir() else (p.suffix != ".py")
        ) or (not path_filter(p)):
            continue

        p = importlib.import_module(f".{p.stem}", package)
        assert p
        modules.append(p)

    return modules


if sys.version_info >= (3, 11):
    from enum import StrEnum as StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        @override
        @staticmethod
        def _generate_next_value_(
            name: str,
            start: int,
            count: int,
            last_values: list[str],
        ):
            return name.lower()


# https://stackoverflow.com/questions/71968447/python-typing-copy-kwargs-from-one-function-to-another
def copy_func_annotations(_source: Callable[P, T]):
    def deco(func: Callable[..., T]) -> Callable[P, T]:
        return func

    return deco


def copy_func_arg_annotations(_source: Callable[P, Any]):
    def deco(func: Callable[..., T]) -> Callable[P, T]:
        return func

    return deco



from collections.abc import Iterator
from contextlib import contextmanager
from functools import partial
from typing import Union

from loguru import logger

from ..common import LazyGetterType, lazy_get


@contextmanager
def logged_suppress(
    msg: "LazyGetterType[str, [Exception]]",
    *t: type[Exception],
    level: Union[int, str] = "ERROR",
    log_stack: bool = True,
    debug_stack: bool = False,
    append_exc_msg: bool = False,
) -> Iterator[None]:
    try:
        yield
    except Exception as e:
        if t and (not isinstance(e, t)):
            raise

        msg_str = lazy_get(msg, e)
        if append_exc_msg:
            msg_str += f": {type(e).__name__}: {e}"

        lg = logger.opt(exception=e if log_stack else False, depth=2)
        lg.log(level, msg_str)

        if debug_stack:
            logger.opt(exception=e, depth=2).debug("Stacktrace")
        return


warning_suppress = partial(
    logged_suppress,
    level="WARNING",
    log_stack=False,
    debug_stack=True,
    append_exc_msg=True,
)



import json
from pathlib import Path


def test_debug_file_writer(tmp_path: Path):
    from cookit import DebugFileWriter

    # Test enabled states
    writer = DebugFileWriter(tmp_path, "subdir")
    assert writer.enabled is True

    non_existent = tmp_path / "nonexistent"
    writer_disabled = DebugFileWriter(non_existent, "subdir")
    assert writer_disabled.enabled is False

    # Test different content types
    content_bytes = b"bytes content"
    writer.write(content_bytes, "test.bin")
    assert (tmp_path / "subdir/test.bin").read_bytes() == content_bytes

    content_str = "text content"
    writer.write(content_str, "test.txt")
    assert (tmp_path / "subdir/test.txt").read_text("utf-8") == content_str

    content_json = {"key": "value"}
    writer.write(content_json, "test.json")
    assert json.loads((tmp_path / "subdir/test.json").read_text()) == content_json

    # Test directory creation
    nested_writer = DebugFileWriter(tmp_path, "subdir", "nested")
    nested_writer.write("content", "test.txt")
    assert (tmp_path / "subdir/nested/test.txt").exists()

    # Test timestamp in filename
    root_writer = DebugFileWriter(tmp_path)
    root_writer.write("content", "test_{time}.txt")
    files = list(tmp_path.glob("test_*.txt"))
    assert len(files) == 1
    assert files[0].read_text() == "content"



from contextlib import asynccontextmanager
from typing import Optional

from nonebot import logger
from nonebot.exception import NoneBotException
from nonebot.matcher import current_matcher


@asynccontextmanager
async def exception_notify(
    msg: str,
    log_msg: Optional[str] = None,
    types: Optional[tuple[type[BaseException]]] = None,
    ignore_nb_exc: bool = True,
):
    try:
        yield
    except Exception as e:
        if ignore_nb_exc and isinstance(e, NoneBotException):
            return
        if types and (not isinstance(e, types)):
            raise
        msg = msg.format(e=str(e), type=type(e).__name__)
        log_msg = log_msg.format(e=str(e), type=type(e).__name__) if log_msg else msg
        logger.opt(depth=2).exception(log_msg)
        await current_matcher.get().finish(msg)
    return  # noqa: PLR1711



from typing import TYPE_CHECKING

import pytest
from nonebot import on_message
from nonebot.exception import FinishedException, SkippedException
from nonebot.matcher import Matcher

from .utils import FakeMessage, make_fake_event

if TYPE_CHECKING:
    from nonebug import App


async def test_exception_notify(app: "App"):
    from cookit.nonebot.base.tools import exception_notify

    cmd_ev = make_fake_event(_message=FakeMessage("114514"))()

    matcher1 = on_message()

    @matcher1.handle()
    async def _():
        async with exception_notify("err"):
            raise ValueError("test")

    async with app.test_matcher(matcher1) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, cmd_ev)
        ctx.should_call_send(cmd_ev, "err")

    matcher2 = on_message()

    @matcher2.handle()
    async def _(m: Matcher):
        async with exception_notify("err"):
            await m.finish("finished")

    async with app.test_matcher(matcher2) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, cmd_ev)
        ctx.should_call_send(cmd_ev, "finished")
        ctx.should_finished()

    matcher3 = on_message()

    @matcher3.handle()
    async def _():
        async with exception_notify("err", types=(ValueError,)):
            raise ValueError("test")

    async with app.test_matcher(matcher3) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, cmd_ev)
        ctx.should_call_send(cmd_ev, "err")

    matcher4 = on_message()

    @matcher4.handle()
    async def _(m: Matcher):
        with pytest.raises(IndexError):
            async with exception_notify("err", types=(ValueError,)):
                raise IndexError("0")
        await m.finish("err expected")

    async with app.test_matcher(matcher4) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, cmd_ev)
        ctx.should_call_send(cmd_ev, "err expected")
        ctx.should_finished()

    matcher5 = on_message()

    @matcher5.handle()
    async def _(m: Matcher):
        async with exception_notify("err", ignore_nb_exc=False):
            await m.finish("finished")

    async with app.test_matcher(matcher5) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, cmd_ev)
        ctx.should_call_send(cmd_ev, "finished")
        ctx.should_call_send(cmd_ev, "err")

    matcher6 = on_message()

    @matcher6.handle()
    async def _(m: Matcher):
        async with exception_notify(
            "err",
            types=(FinishedException,),
            ignore_nb_exc=False,
        ):
            await m.finish("finished")

    async with app.test_matcher(matcher6) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, cmd_ev)
        ctx.should_call_send(cmd_ev, "finished")
        ctx.should_call_send(cmd_ev, "err")

    matcher7 = on_message()

    @matcher7.handle()
    async def _(m: Matcher):
        async with exception_notify(
            "err",
            types=(SkippedException,),
            ignore_nb_exc=False,
        ):
            await m.finish("finished")

    async with app.test_matcher(matcher7) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, cmd_ev)
        ctx.should_call_send(cmd_ev, "finished")
        ctx.should_finished()



import mimetypes
import re
from collections.abc import Awaitable, Iterable
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Optional,
    Protocol,
    TypedDict,
    TypeVar,
    Union,
)
from typing_extensions import TypeAlias, Unpack

import anyio
from playwright.async_api import Page, Request, Route
from yarl import URL

if TYPE_CHECKING:
    from pathlib import Path

PWRouter: TypeAlias = Callable[[Route, Request], Awaitable[Any]]
CKRouterPattern: TypeAlias = Union[str, re.Pattern]

T_co = TypeVar("T_co", covariant=True)
TF = TypeVar("TF", bound="CKRouterFunc")


class CKRouterKwArgs(TypedDict):
    route: Route
    request: Request
    info: "CKRouterInfo"
    url: URL
    matched: Optional[re.Match[str]]


class CKRouterFunc(Protocol, Generic[T_co]):
    def __call__(self, **kwds: Unpack[CKRouterKwArgs]) -> Awaitable[T_co]: ...


@dataclass
class CKRouterInfo:
    pattern: CKRouterPattern
    func: CKRouterFunc[Any]
    priority: int


async def apply_router_to_page(page: Page, router: CKRouterInfo):
    async def wrapped(route: Route, request: Request):
        url = URL(request.url)
        matched = (
            re.search(router.pattern, request.url)
            if isinstance(router.pattern, re.Pattern)
            else None
        )
        return await router.func(
            route=route,
            request=request,
            info=router,
            url=url,
            matched=matched,
        )

    await page.route(router.pattern, wrapped)


class RouterGroup:
    def __init__(self, routers: Optional[Iterable[CKRouterInfo]] = None) -> None:
        self.routers: list[CKRouterInfo] = []
        if routers:
            self.routers.extend(routers)

    def register_router(
        self,
        func: CKRouterFunc,
        pattern: CKRouterPattern,
        priority: int = 1,
    ):
        self.routers.append(CKRouterInfo(pattern, func, priority))
        # 低 priority 的 CKRouter 应最先运行，
        # 因为 playwright 后 route 的先运行，所以要反过来排序
        self.routers.sort(key=lambda r: r.priority, reverse=True)

    def router(
        self,
        pattern: CKRouterPattern,
        priority: int = 1,
    ):
        def wrapper(func: TF) -> TF:
            self.register_router(func, pattern, priority)
            return func

        return wrapper

    async def apply(self, page: Page):
        for router in self.routers:
            await apply_router_to_page(page, router)
        return page

    def copy(self) -> "RouterGroup":
        return RouterGroup(self.routers)


def make_real_path_router(path_extractor: CKRouterFunc["Path"]) -> CKRouterFunc:
    async def router(route: Route, matched: Optional[re.Match[str]], **add_kwds):
        path = await path_extractor(route=route, matched=matched, **add_kwds)

        if (not path.exists()) or (not path.is_file()):
            return await route.fulfill(status=404)

        body = await anyio.Path(path).read_bytes()
        mime = mimetypes.guess_type(path)[0] or "application/octet-stream"
        return await route.fulfill(body=body, content_type=mime)

    return router



from typing import TYPE_CHECKING

from .utils import FakeMessage, make_fake_event

if TYPE_CHECKING:
    from nonebug import App


async def test_command_arg_plaintext(app: "App"):
    from cookit.nonebot import CommandArgPlaintext
    from nonebot import on_command
    from nonebot.matcher import Matcher

    ev = make_fake_event(_message=FakeMessage("/test resp1\n"))()
    ev_empty = make_fake_event(_message=FakeMessage("/test"))()

    matcher1 = on_command("test")

    @matcher1.handle()
    async def _(m: Matcher, arg: str = CommandArgPlaintext()):
        await m.finish(arg)

    async with app.test_matcher(matcher1) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, ev)
        ctx.should_call_send(ev, "resp1")
        ctx.receive_event(bot, ev_empty)
        ctx.should_call_send(ev_empty, "")

    matcher2 = on_command("test")

    @matcher2.handle()
    async def _(m: Matcher, arg: str = CommandArgPlaintext(strip=False)):
        await m.finish(arg)

    async with app.test_matcher(matcher2) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, ev)
        ctx.should_call_send(ev, "resp1\n")
        ctx.receive_event(bot, ev_empty)
        ctx.should_call_send(ev_empty, "")

    matcher3 = on_command("test")

    @matcher3.handle()
    async def _(m: Matcher, arg: str = CommandArgPlaintext(allow_empty=False)):
        await m.finish(arg)

    @matcher3.handle()
    async def _(m: Matcher):
        await m.finish("empty arg")

    async with app.test_matcher(matcher3) as ctx:
        bot = ctx.create_bot()
        ctx.receive_event(bot, ev)
        ctx.should_call_send(ev, "resp1")
        ctx.receive_event(bot, ev_empty)
        ctx.should_call_send(ev_empty, "empty arg")



from .dependency import (
    CommandArgPlaintext as CommandArgPlaintext,
    dep_command_arg_plaintext as dep_command_arg_plaintext,
)
from .plugin import (
    assert_plugin_loaded as assert_plugin_loaded,
)
from .tools import (
    exception_notify as exception_notify,
)



from collections.abc import MutableMapping
from functools import partial
from typing import Any, Callable, Generic, Protocol, TypeVar, Union, overload
from typing_extensions import deprecated

from cookit.common import HasNameProtocol

T = TypeVar("T")
T_HasName = TypeVar("T_HasName", bound=HasNameProtocol)


@overload
def append_obj_to_dict_deco(
    name_dict: MutableMapping[str, T_HasName],
    obj_or_name: T_HasName,
    allow_overwrite: bool = True,
) -> T_HasName: ...
@overload
def append_obj_to_dict_deco(
    name_dict: MutableMapping[str, T],
    obj_or_name: str,
    allow_overwrite: bool = True,
) -> Callable[[T], T]: ...
@deprecated("Use `NameDecoCollector` instead")
def append_obj_to_dict_deco(  # type: ignore
    name_dict: MutableMapping[str, T],
    obj_or_name: Union[str, T],
    allow_overwrite: bool = True,
):
    def set_obj(name: str, obj: T) -> T:
        if (not allow_overwrite) and (name in name_dict):
            raise ValueError(f"Object with name '{name}' already exists")
        name_dict[name] = obj
        return obj

    if isinstance(obj_or_name, str):

        def inner_deco(obj: T, /):
            return set_obj(obj_or_name, obj)

        return inner_deco

    if isinstance(
        (name := getattr(obj_or_name, "__name__", None)),
        str,
    ):
        return set_obj(name, obj_or_name)

    raise TypeError("func_or_name must be str or object with __name__ attribute")


class AppendObjDecoProtocol(Protocol, Generic[T]):
    @overload
    def __call__(self, x: T) -> T: ...
    @overload
    def __call__(self, x: str) -> Callable[[T], T]: ...


@overload
def make_append_obj_to_dict_deco(
    name_dict: MutableMapping[str, T_HasName],
    allow_overwrite: bool = True,
) -> AppendObjDecoProtocol[T_HasName]: ...
@overload
def make_append_obj_to_dict_deco(
    name_dict: MutableMapping[str, T],
    allow_overwrite: bool = True,
) -> Callable[[str], Callable[[T], T]]: ...
@deprecated("Use `NameDecoCollector` instead")
def make_append_obj_to_dict_deco(name_dict: MutableMapping[str, Any], **kwargs):  # type: ignore
    return partial(append_obj_to_dict_deco, name_dict, **kwargs)


@overload
def append_func_to_dict_deco(
    name_dict: MutableMapping[str, T_HasName],
    func_or_name: T_HasName,
) -> T_HasName: ...
@overload
def append_func_to_dict_deco(
    name_dict: MutableMapping[str, T],
    func_or_name: str,
) -> Callable[[T], T]: ...
@deprecated("Use `NameDecoCollector` instead")
def append_func_to_dict_deco(name_dict: Any, func_or_name: Any):
    return append_obj_to_dict_deco(name_dict, func_or_name)


@deprecated("Use `NameDecoCollector` instead")
def make_append_func_to_dict_deco(name_dict: dict[str, Callable]):
    return make_append_obj_to_dict_deco(name_dict)



import sys
from typing import Any, Generic, TypeVar, overload
from typing_extensions import override

T = TypeVar("T")

if sys.version_info >= (3, 10):
    from contextlib import nullcontext as nullcontext

else:
    from contextlib import AbstractAsyncContextManager, AbstractContextManager

    class nullcontext(AbstractContextManager, AbstractAsyncContextManager, Generic[T]):  # noqa: N801
        @overload
        def __init__(self: "nullcontext[None]") -> None: ...
        @overload
        def __init__(self, enter_result: T) -> None: ...
        def __init__(self, enter_result: Any = None):
            super().__init__()
            self.enter_result: T = enter_result

        @override
        def __enter__(self):
            return self.enter_result

        @override
        def __exit__(self, *_):
            pass

        @override
        async def __aenter__(self):
            return self.enter_result

        @override
        async def __aexit__(self, *_):
            pass



from typing import Any

from nonebot.adapters import Message as BaseMessage
from nonebot.matcher import Matcher
from nonebot.params import CommandArg, Depends

from ... import copy_func_arg_annotations


def dep_command_arg_plaintext(strip: bool = True, allow_empty: bool = True):
    async def dep(arg: BaseMessage = CommandArg()) -> str:
        t = arg.extract_plain_text()
        if strip:
            t = t.strip()
        if not allow_empty and not t:
            Matcher.skip()
        return t

    return dep


@copy_func_arg_annotations(dep_command_arg_plaintext)
def CommandArgPlaintext(*args, **kwargs) -> Any:  # noqa: N802
    return Depends(dep_command_arg_plaintext(*args, **kwargs))



from typing import Callable, cast

import pytest


def test_lazy_get():
    from cookit import LazyGetterType, lazy_get

    a = cast("LazyGetterType[int, []]", 5)
    result = lazy_get(a)
    assert result == a

    b = lambda: 10  # noqa: E731
    result = lazy_get(b)
    assert result == b()

    c = cast("Callable[[int], int]", lambda x: x + 1)  # noqa: E731
    result = lazy_get(c, 1)
    assert result == c(1)


def test_qor_with_a_not_none():
    from cookit import qor

    a = 5
    b = 10
    result = qor(a, b)
    assert result == a


def test_qor_with_a_none_and_b_not_callable():
    from cookit import qor

    a = None
    b = 10
    result = qor(a, b)
    assert result == b


def test_qor_with_a_none_and_b_callable():
    from cookit import qor

    a = None
    b = lambda: 10  # noqa: E731
    result = qor(a, b)
    assert result == b()


def test_chunks():
    from cookit import chunks

    lst = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    assert list(chunks(lst, 3)) == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    lst2 = [1, 2, 3, 4, 5, 6, 7, 8]
    assert list(chunks(lst2, 3)) == [[1, 2, 3], [4, 5, 6], [7, 8]]


def test_flatten():
    from cookit import flatten

    lst = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    assert flatten(lst) == [1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_set_default_with_existing_key():
    from cookit import set_default

    target = {"a": 1, "b": 2}
    key = "a"
    default = 10
    result = set_default(target, key, default)
    assert result == target[key]
    assert len(target) == 2


def test_set_default_with_new_key():
    from cookit import set_default

    target = {"a": 1, "b": 2}
    key = "c"
    default = 10
    result = set_default(target, key, default)
    assert result == default
    assert target[key] == default
    assert len(target) == 3


def test_set_default_with_callable_default():
    from cookit import set_default

    target = {"a": 1, "b": 2}
    key = "c"
    default = lambda: 10  # noqa: E731
    result = set_default(target, key, default)
    assert result == default()
    assert target[key] == default()
    assert len(target) == 3


def test_auto_delete():
    from cookit import auto_delete

    # Test case 1: No transformation function provided
    target1 = {"a": 1, "b": 2, "c": None, "d": 0}
    expected_result1 = {"a": 1, "b": 2}
    result1 = auto_delete(target1)
    assert result1 == expected_result1
    assert target1 == expected_result1

    # Test case 2: Transformation function provided
    def transform_func(value: int):
        return value * 2 if value % 2 == 0 else None

    target2 = {"a": 1, "b": 2, "c": 3, "d": 4}
    expected_result2 = {"b": 2, "d": 4}
    expected_result2_transformed = {"b": 4, "d": 8}
    result2 = auto_delete(target2, transform=transform_func)
    assert result2 == expected_result2_transformed
    assert target2 == expected_result2

    # Test case 3: Empty target dictionary
    target3 = {}
    expected_result3 = {}
    result3 = auto_delete(target3)
    assert result3 == expected_result3
    assert target3 == expected_result3


def test_deco_collector():
    from cookit import DecoCollector

    collector = DecoCollector[str, int]()
    collector("test1")(1)
    assert collector.data == {"test1": 1}
    collector("test2")(2)
    assert collector.data == {"test1": 1, "test2": 2}
    with pytest.raises(ValueError, match="Object with key 'test1' already exists"):
        collector("test1")(114514)

    collector2_data: dict[str, int] = {}
    collector2 = DecoCollector(collector2_data, allow_overwrite=True)
    collector2("test2_1")(1)
    assert collector2.data == {"test2_1": 1}
    assert collector2.data is collector2_data
    collector2("test2_1")(114514)
    assert collector2.data == {"test2_1": 114514}


def test_type_deco_collector():
    from cookit import TypeDecoCollector

    class BaseCls:
        pass

    class Cls1(BaseCls):
        pass

    class Cls2(BaseCls):
        pass

    collector = TypeDecoCollector[BaseCls, int]()
    collector(Cls1)(1)
    assert collector.data == {Cls1: 1}
    collector(Cls2)(2)
    assert collector.data == {Cls1: 1, Cls2: 2}
    with pytest.raises(
        ValueError,
        match="Object with key '<class '.+Cls1'>' already exists",
    ):
        collector(Cls1)(114514)


def test_name_deco_collector():
    from cookit import NameDecoCollector

    class Cls1:
        pass

    class Cls2:
        pass

    collector = NameDecoCollector[type]()
    collector(Cls1)
    assert collector.data == {"Cls1": Cls1}
    with pytest.raises(ValueError, match="Object with key 'Cls1' already exists"):
        collector(Cls1)
    collector("ClsABC")(Cls2)
    assert collector.data == {"Cls1": Cls1, "ClsABC": Cls2}
    with pytest.raises(ValueError, match="Object with key 'Cls1' already exists"):
        collector("Cls1")(Cls2)


def test_deep_merge():
    from cookit import deep_merge

    assert deep_merge(1, 2, 3) == 3
    assert deep_merge([1, 2], [3, 4], [5]) == [1, 2, 3, 4, 5]
    assert deep_merge([1, 2], [3, 4], 5) == 5
    assert deep_merge({"a": 1, "b": 2}, {"a": 3, "c": 4}) == {"a": 3, "b": 2, "c": 4}
    assert deep_merge(
        {"a": 1, "b": 2},
        {"a": 3, "c": 4},
        {"b": 3},
    ) == {"a": 3, "b": 3, "c": 4}
    assert deep_merge(
        {"a": [1, 2], "b": [3, 4]},
        {"a": [5, 6], "c": [7, 8]},
    ) == {"a": [1, 2, 5, 6], "b": [3, 4], "c": [7, 8]}
    assert deep_merge(
        {"a": [1, 2], "b": [3, 4]},
        {"a": [5, 6], "b": 6},
    ) == {"a": [1, 2, 5, 6], "b": 6}

    assert deep_merge(
        {"a": 1},
        {"b": [2, 3]},
        {"b": [4, 5]},
    ) == {"a": 1, "b": [2, 3, 4, 5]}
    assert deep_merge(
        {"a": 1},
        {"b": [2, 3]},
        {"b": [4, 5]},
        skip_merge_paths={"b"},
    ) == {"a": 1, "b": [4, 5]}

    assert deep_merge(
        {"a": {"b": {"c": [1, 2]}}},
        {"a": {"b": {"c": [3, 4]}}},
    ) == {"a": {"b": {"c": [1, 2, 3, 4]}}}
    assert deep_merge(
        {"a": {"b": {"c": [1, 2]}}},
        {"a": {"b": {"c": [3, 4]}}},
        skip_merge_paths={"a.b.c"},
    ) == {"a": {"b": {"c": [3, 4]}}}

    assert deep_merge((1, 2), (3, 4)) == (3, 4)
    assert deep_merge({1, 2}, {3, 4}) == {1, 2, 3, 4}

    assert deep_merge((1, 2), {3, 4}) == {3, 4}
    assert deep_merge([1, 2], (3, 4)) == (3, 4)



from .data import (
    AppendObjDecoProtocol as AppendObjDecoProtocol,
    append_func_to_dict_deco as append_func_to_dict_deco,
    append_obj_to_dict_deco as append_obj_to_dict_deco,
    make_append_func_to_dict_deco as make_append_func_to_dict_deco,
    make_append_obj_to_dict_deco as make_append_obj_to_dict_deco,
)
from .sized_list import SizedList as SizedList



import asyncio
from pathlib import Path


async def test_file_cache_manager(tmp_path: Path):
    from cookit import FileCacheManager

    bytes_cache = FileCacheManager(tmp_path, max_size=2, ttl=1)

    bytes_cache["test"] = b"test"
    assert "test" in bytes_cache
    assert bytes_cache.get("test") == b"test"

    bytes_cache["test"] = b"test test"
    assert "test" in bytes_cache
    assert bytes_cache.get("test") == b"test test"

    await asyncio.sleep(0.01)
    bytes_cache["test2"] = b"test2"
    assert "test2" in bytes_cache
    assert bytes_cache.get("test2") == b"test2"

    await asyncio.sleep(0.01)
    bytes_cache["test3"] = b"test3"
    assert "test3" in bytes_cache
    assert bytes_cache.get("test3") == b"test3"

    assert "test2" in bytes_cache
    assert bytes_cache.get("test2") == b"test2"

    assert "test" not in bytes_cache
    assert bytes_cache.get("test") is None

    await asyncio.sleep(0.5)

    await asyncio.sleep(0.01)
    bytes_cache["test4"] = b"test4"
    assert "test4" in bytes_cache
    assert bytes_cache.get("test4") == b"test4"

    assert "test" not in bytes_cache
    assert bytes_cache.get("test") is None
    assert "test2" not in bytes_cache
    assert bytes_cache.get("test2") is None

    await asyncio.sleep(0.5)

    assert "test" not in bytes_cache
    assert bytes_cache.get("test") is None
    assert "test2" not in bytes_cache
    assert bytes_cache.get("test2") is None
    assert "test3" not in bytes_cache
    assert bytes_cache.get("test3") is None
    assert "test4" in bytes_cache
    assert bytes_cache.get("test4") == b"test4"

    await asyncio.sleep(0.5)

    assert "test" not in bytes_cache
    assert bytes_cache.get("test") is None
    assert "test2" not in bytes_cache
    assert bytes_cache.get("test2") is None
    assert "test3" not in bytes_cache
    assert bytes_cache.get("test3") is None
    assert "test4" not in bytes_cache
    assert bytes_cache.get("test4") is None

    text_cache = FileCacheManager(tmp_path, text_mode=True, max_size=2, ttl=1)

    text_cache["test_text"] = "test测试测试"
    assert "test_text" in text_cache
    assert text_cache.get("test_text") == "test测试测试"
    assert bytes_cache.get("test_text") == "test测试测试".encode()



from typing import TYPE_CHECKING, Callable

from ..common import NameDecoCollector

if TYPE_CHECKING:
    from jinja2 import Environment


def make_register_jinja_filter_deco(env: "Environment"):
    return NameDecoCollector[Callable](env.filters)



async def test_async_nullcontext():
    from cookit.common.compat import nullcontext

    async with nullcontext() as value:
        assert value is None

    async with nullcontext(42) as value:
        assert value == 42



from typing import Callable, Generic, Optional, Protocol, TypeVar, Union, overload
from typing_extensions import override

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class HasNameProtocol(Protocol):
    __name__: str


T_HasName = TypeVar("T_HasName", bound=HasNameProtocol)


class DecoCollector(Generic[K, V]):
    def __init__(
        self,
        data: Optional[dict[K, V]] = None,
        allow_overwrite: bool = False,
    ) -> None:
        super().__init__()
        self.data: dict[K, V] = {} if data is None else data
        self.allow_overwrite = allow_overwrite

    def set_data_item(self, key: K, value: V, /) -> None:
        if (not self.allow_overwrite) and (key in self.data):
            raise ValueError(f"Object with key '{key}' already exists")
        self.data[key] = value

    def __call__(self, key: K) -> Callable[[V], V]:
        def deco(obj: V) -> V:
            self.set_data_item(key, obj)
            return obj

        return deco


class TypeDecoCollector(DecoCollector[type[K], V]):
    @overload
    def get_from_type_or_instance(self, obj: Union[type[K], K]) -> V: ...
    @overload
    def get_from_type_or_instance(
        self,
        obj: Union[type[K], K],
        default: T = ...,
    ) -> Union[V, T]: ...
    def get_from_type_or_instance(
        self,
        obj: Union[type[K], K],
        default: T = ...,
    ) -> Union[V, T]:
        type_key: type[K] = obj if isinstance(obj, type) else type(obj)
        try:
            return self.data[type_key]
        except KeyError:
            if default is ...:
                raise
            return default


class NameDecoCollector(DecoCollector[str, Union[T_HasName]]):
    @overload
    def __call__(self, key: str) -> Callable[[T_HasName], T_HasName]: ...
    @overload
    def __call__(self, key: T_HasName) -> T_HasName: ...
    @override
    def __call__(self, key: Union[str, T_HasName]):  # type: ignore[reportIncompatibleMethodOverride]
        if isinstance(key, str):
            return super().__call__(key)
        if isinstance((name := getattr(key, "__name__", None)), str):
            return super().__call__(name)(key)
        raise TypeError("'key' must be str or object with __name__ attribute")
